from typing import Any, Generator
from urllib.parse import urlparse

from pydantic import field_validator

from anystore.exceptions import DoesNotExist
from anystore.mixins import BaseModel
from anystore.serialize import Mode, from_store, to_store
from anystore.settings import Settings
from anystore.types import Uri, Value, Model
from anystore.util import clean_dict, ensure_uri


settings = Settings()


class BaseStore(BaseModel):
    uri: Uri | None = settings.uri
    scheme: str | None = None
    serialization_mode: Mode | None = settings.serialization_mode
    model: Model | None = None
    raise_on_nonexist: bool | None = settings.raise_on_nonexist
    backend_config: dict[str, Any] | None = None

    def __init__(self, **data):
        uri = data.get("uri", settings.uri)
        data["scheme"] = urlparse(str(uri)).scheme
        super().__init__(**data)

    def _write(self, key: Uri, value: Value, **kwargs) -> None:
        """
        Write value with key to acutal backend
        """
        raise NotImplementedError

    def _read(self, key: Uri, raise_on_nonexist: bool | None = True, **kwargs) -> Any:
        """
        Read key from actual backend
        """
        raise NotImplementedError

    def _stream(self, key: Uri, raise_on_nonexist: bool | None = True, **kwargs) -> Any:
        """
        Stream key line by line from actual backend (for file-like powered backend)
        """
        raise NotImplementedError

    def _exists(self, key: Uri) -> bool:
        """
        Check if the given key exists
        """
        raise NotImplementedError

    def _get_key_prefix(self) -> str:
        """
        Get backend specific key prefix
        """
        raise NotImplementedError

    def _iterate_keys(self, prefix: str | None = None) -> Generator[str, None, None]:
        """
        Backend specific key iterator
        """
        raise NotImplementedError

    def get(
        self,
        key: Uri,
        raise_on_nonexist: bool | None = None,
        serialization_mode: Mode | None = None,
        model: Model | None = None,
        **kwargs,
    ) -> Any:
        serialization_mode = serialization_mode or self.serialization_mode
        model = model or self.model
        if raise_on_nonexist is None:
            raise_on_nonexist = self.raise_on_nonexist
        kwargs = self.ensure_kwargs(**kwargs)
        key = self.get_key(key)
        try:
            return from_store(
                self._read(key, raise_on_nonexist, **kwargs),
                serialization_mode,
                model=model,
            )
        except FileNotFoundError:  # fsspec
            if raise_on_nonexist:
                raise DoesNotExist(f"Key does not exist: `{key}`")
            return None

    def stream(
        self,
        key: Uri,
        raise_on_nonexist: bool | None = None,
        serialization_mode: Mode | None = None,
        model: Model | None = None,
        **kwargs,
    ) -> Generator[Any, None, None]:
        key = self.get_key(key)
        model = model or self.model
        try:
            for line in self._stream(key, raise_on_nonexist, **kwargs):
                yield from_store(line, serialization_mode, model=model)
        except FileNotFoundError:  # fsspec
            if raise_on_nonexist:
                raise DoesNotExist(f"Key does not exist: `{key}`")
            return None

    def put(
        self,
        key: Uri,
        value: Any,
        serialization_mode: Mode | None = None,
        model: Model | None = None,
        **kwargs,
    ):
        serialization_mode = serialization_mode or self.serialization_mode
        model = model or self.model
        kwargs = self.ensure_kwargs(**kwargs)
        key = self.get_key(key)
        self._write(key, to_store(value, serialization_mode, model=model))

    def exists(self, key: Uri) -> bool:
        return self._exists(key)

    def ensure_kwargs(self, **kwargs) -> dict[str, Any]:
        config = clean_dict(self.backend_config)
        return {**config, **clean_dict(kwargs)}

    def get_key(self, key: Uri) -> str:
        return f"{self._get_key_prefix()}/{str(key)}".strip("/")

    def iterate_keys(self, prefix: str | None = None) -> Generator[str, None, None]:
        yield from self._iterate_keys(prefix)

    @field_validator("uri", mode="before")
    @classmethod
    def ensure_uri(cls, v: Any) -> str:
        uri = ensure_uri(v)
        return uri.rstrip("/")



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from anystore.serialize import Mode


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="anystore_")

    uri: str | None = ".anystore"
    yaml_uri: str | None = None
    json_uri: str | None = None
    serialization_mode: Mode | None = "auto"
    raise_on_nonexist: bool = True

    debug: bool = Field(alias="debug", default=False)
    redis_debug: bool = Field(alias="redis_debug", default=False)


class SqlSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="anystore_sql_")

    table: str | None = "anystore"
    pool_size: int | None = 5



from io import BytesIO
import hashlib
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from banal import clean_dict as _clean_dict
from banal import is_mapping


def _clean(val: Any) -> Any:
    if val is False:
        return False
    return val or None


def clean_dict(data: Any) -> dict[str, Any]:
    """
    strip out defaultdict and ensure str keys (for serialization)
    """
    if not is_mapping(data):
        return {}
    return _clean_dict(
        {
            str(k): clean_dict(dict(v)) or None if is_mapping(v) else _clean(v)
            for k, v in data.items()
        }
    )


def ensure_uri(uri: Any) -> str:
    if not uri:
        raise ValueError(f"Invalid uri: `{uri}`")
    if uri == "-":  # stdin/stoud
        return uri
    if isinstance(uri, Path):
        return uri.absolute().as_uri()
    if isinstance(uri, str) and not uri.strip():
        raise ValueError(f"Invalid uri: `{uri}`")
    uri = str(uri)
    parsed = urlparse(uri)
    if parsed.scheme:
        return uri
    return Path(uri).absolute().as_uri()


def make_checksum(io: BytesIO, algorithm: str = "md5") -> str:
    hash_ = getattr(hashlib, algorithm)()
    for chunk in iter(lambda: io.read(128 * hash_.block_size), b""):
        hash_.update(chunk)
    return hash_.hexdigest()


def make_data_checksum(data: Any, algorithm: str = "md5") -> str:
    data = repr(data).encode()
    return make_checksum(BytesIO(data), algorithm)


def make_signature_key(*args, **kwargs) -> str:
    return make_data_checksum((args, kwargs))



import contextlib
import logging
from pathlib import Path
import sys
from typing import Any, BinaryIO, Generator, TextIO, TypeAlias

from fsspec import open
from fsspec.core import OpenFile

from anystore.util import ensure_uri

log = logging.getLogger(__name__)

DEFAULT_MODE = "rb"
DEFAULT_WRITE_MODE = "wb"

Uri: TypeAlias = Path | BinaryIO | TextIO | str


def _get_sysio(mode: str | None = DEFAULT_MODE) -> TextIO | BinaryIO:
    if mode and mode.startswith("r"):
        return sys.stdin
    return sys.stdout


class SmartHandler:
    def __init__(
        self,
        uri: Uri,
        **kwargs,
    ) -> None:
        self.uri = ensure_uri(uri)
        self.is_buffer = self.uri == "-"
        kwargs["mode"] = kwargs.get("mode", DEFAULT_MODE)
        self.sys_io = _get_sysio(kwargs["mode"])
        if hasattr(self.sys_io, "buffer"):
            self.sys_io = self.sys_io.buffer
        self.kwargs = kwargs
        self.handler: OpenFile | TextIO | None = None

    def open(self):
        if self.is_buffer:
            self.handler = self.sys_io
        else:
            handler = open(self.uri, **self.kwargs)
            self.handler = handler.open()
        return self.handler

    def close(self):
        if not self.is_buffer and self.handler is not None:
            self.handler.close()

    def __enter__(self):
        return self.open()

    def __exit__(self, *args, **kwargs) -> None:
        self.close()


@contextlib.contextmanager
def smart_open(
    uri: Uri,
    mode: str | None = DEFAULT_MODE,
    **kwargs,
):
    handler = SmartHandler(uri, mode=mode, **kwargs)
    try:
        yield handler.open()
    finally:
        handler.close()


def smart_stream(
    uri: Uri, mode: str | None = DEFAULT_MODE, **kwargs
) -> Generator[str | bytes, None, None]:
    with smart_open(uri, mode, **kwargs) as fh:
        while line := fh.readline():
            yield line


def smart_read(uri: Uri, mode: str | None = DEFAULT_MODE, **kwargs) -> Any:
    with smart_open(uri, mode, **kwargs) as fh:
        return fh.read()


def smart_write(
    uri: Uri, content: bytes | str, mode: str | None = DEFAULT_WRITE_MODE, **kwargs
) -> None:
    if uri == "-":
        if isinstance(content, str):
            content = content.encode()
    with smart_open(uri, mode, **kwargs) as fh:
        fh.write(content)



from functools import cache
from logging import getLogger
from urllib.parse import urlparse

from anystore.settings import Settings
from anystore.store.base import BaseStore
from anystore.store.fs import Store
from anystore.util import ensure_uri


log = getLogger(__name__)


@cache
def get_store(settings: Settings | None = None, **kwargs) -> BaseStore:
    settings = settings or Settings()
    uri = kwargs.get("uri")
    if uri is None:
        if settings.yaml_uri is not None:
            store = BaseStore.from_yaml_uri(settings.yaml_uri, **kwargs)
            return get_store(**store.model_dump())
        if settings.json_uri is not None:
            store = BaseStore.from_json_uri(settings.json_uri, **kwargs)
            return get_store(**store.model_dump())
        uri = settings.uri
    uri = ensure_uri(uri)
    parsed = urlparse(uri)
    if parsed.scheme == "redis":
        try:
            from anystore.store.redis import RedisStore

            return RedisStore(**kwargs)
        except ImportError:
            log.error("Install redis dependencies via `anystore[redis]`")
    if "sql" in parsed.scheme:
        try:
            from anystore.store.sql import SqlStore

            return SqlStore(**kwargs)
        except ImportError:
            log.error("Install sql dependencies via `anystore[sql]`")
    return Store(**kwargs)


__all__ = ["get_store", "Store", "RedisStore", "SqlStore"]



from functools import lru_cache
import logging
from typing import Self

import orjson
from pydantic import BaseModel as _BaseModel
from pydantic import field_validator
import yaml

from anystore.io import smart_read
from anystore.types import Uri
from anystore.util import clean_dict

log = logging.getLogger(__name__)


@lru_cache(128)
def cached_from_uri(uri: Uri) -> str:
    """
    Cache remote sources on runtime
    """
    log.info("Loading `%s` ..." % uri)
    return smart_read(uri)


class JsonMixin:
    """
    Load a pydantic model from a local or remote uri with json data
    """

    @classmethod
    def from_json_str(cls, data: str, **kwargs) -> Self:
        loaded = clean_dict({**orjson.loads(data), **clean_dict(kwargs)})
        return cls(**loaded)

    @classmethod
    def from_json_uri(cls, uri: Uri, **kwargs) -> Self:
        data = cached_from_uri(uri)
        return cls.from_json_str(data, **kwargs)


class YamlMixin:
    """
    Load a pydantic model from a local or remote uri with yaml data
    """

    @classmethod
    def from_yaml_str(cls, data: str, **kwargs) -> Self:
        loaded = clean_dict({**yaml.safe_load(data), **clean_dict(kwargs)})
        return cls(**loaded)

    @classmethod
    def from_yaml_uri(cls, uri: Uri, **kwargs) -> Self:
        data = cached_from_uri(uri)
        return cls.from_yaml_str(data, **kwargs)


class RemoteMixin(JsonMixin, YamlMixin):
    @classmethod
    def _from_uri(cls, uri: Uri, **kwargs) -> Self:
        try:
            return cls.from_json_uri(uri, **kwargs)
        except orjson.JSONDecodeError:
            return cls.from_yaml_uri(uri, **kwargs)


class BaseModel(_BaseModel, RemoteMixin):
    def __hash__(self) -> int:
        return hash(repr(self.model_dump()))

    @field_validator("*", mode="before")
    @classmethod
    def empty_str_to_none(cls, v) -> str | None:
        if v == "":
            return None
        return v



import functools
from typing import Callable
from anystore.exceptions import DoesNotExist
from anystore.store import get_store
from anystore.util import make_signature_key


def anycache(func=None, **store_kwargs):
    key_func: Callable = store_kwargs.pop("key_func", make_signature_key)
    serialize_func: Callable = store_kwargs.pop("serialize_func", None)
    store = store_kwargs.pop("store", get_store(**store_kwargs))
    store = store.model_copy()
    store.raise_on_nonexist = True

    def _decorator(func):
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            key = key_func(*args, **kwargs)
            try:
                if key is not None:
                    return store.get(key)
                raise DoesNotExist
            except DoesNotExist:
                res = func(*args, **kwargs)
                if serialize_func is not None:
                    res = serialize_func(res)
                if key is not None:
                    store.put(key, res)
                return res

        return _inner

    if func is None:
        return _decorator
    return _decorator(func)



"""
Store backend using redis-like stores such as Redis, Fakeredis or Apache Kvrocks
"""

from typing import Any, Generator, TYPE_CHECKING
import logging
from functools import cache

import redis

from anystore.exceptions import DoesNotExist
from anystore.settings import Settings
from anystore.store.base import BaseStore
from anystore.types import Uri, Value

if TYPE_CHECKING:
    import fakeredis


log = logging.getLogger(__name__)


@cache
def get_redis(uri: str) -> "fakeredis.FakeStrictRedis | redis.Redis":
    settings = Settings()
    if settings.redis_debug:
        import fakeredis

        con = fakeredis.FakeStrictRedis()
        con.ping()
        log.info("Redis connected: `fakeredis`")
        return con
    con = redis.from_url(uri)
    con.ping()
    log.info("Redis connected: `{uri}`")
    return con


class RedisStore(BaseStore):
    def _write(self, key: Uri, value: Value, **kwargs) -> None:
        con = get_redis(self.uri)
        con.set(str(key), value, **kwargs)

    def _read(self, key: Uri, raise_on_nonexist: bool | None = True, **kwargs) -> Any:
        con = get_redis(self.uri)
        key = str(key)
        # `None` could be stored as an actual value, to implement `raise_on_nonexist`
        # we need to check this first:
        if raise_on_nonexist and not con.exists(key):
            raise DoesNotExist
        res = con.get(str(key))
        # mimic fs read mode:
        if kwargs.get("mode") == "r" and isinstance(res, bytes):
            res = res.decode()
        return res

    def _exists(self, key: Uri) -> bool:
        con = get_redis(self.uri)
        res = con.exists(self.get_key(key))
        return bool(res)

    def _get_key_prefix(self) -> str:
        if self.backend_config is not None:
            return self.backend_config.get("redis_prefix") or "anystore"
        return "anystore"

    def _iterate_keys(self, prefix: str | None = None) -> Generator[str, None, None]:
        con = get_redis(self.uri)
        prefix = self.get_key(prefix or "") + "*"
        key_prefix = self._get_key_prefix()
        for key in con.scan_iter(prefix):
            key = key.decode()
            yield key[len(key_prefix) + 1 :]



from os import PathLike
from pathlib import Path
from typing import Type, TypeAlias

from pydantic import BaseModel


Uri: TypeAlias = PathLike | Path | str
Value: TypeAlias = str | bytes
Model: TypeAlias = Type[BaseModel]



from rich import print
from rich.console import Console
from typing import Annotated, Optional

import typer

from anystore import __version__
from anystore.io import smart_read, smart_write
from anystore.settings import Settings
from anystore.store import get_store

settings = Settings()
cli = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=settings.debug)
console = Console(stderr=True)

state = {"uri": settings.uri, "pickle": False}


class ErrorHandler:
    def __enter__(self):
        pass

    def __exit__(self, e, msg, _):
        if e is not None:
            if settings.debug:
                raise e
            console.print(f"[red][bold]{e.__name__}[/bold]: {msg}[/red]")
            raise typer.Exit(code=1)


@cli.callback(invoke_without_command=True)
def cli_store(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
    store: Annotated[
        Optional[str], typer.Option(..., help="Store base uri")
    ] = settings.uri,
    pickle: Annotated[
        Optional[bool], typer.Option(..., help="Use pickle serializer")
    ] = False,
):
    if version:
        print(__version__)
        raise typer.Exit()
    state["uri"] = store
    state["pickle"] = pickle


@cli.command("get")
def cli_get(
    key: str,
    o: Annotated[str, typer.Option("-o")] = "-",
):
    """
    Get content of a `key` from a store
    """
    with ErrorHandler():
        S = get_store(uri=state["uri"], use_pickle=state["pickle"])
        value = S.get(key)
        mode = "w" if isinstance(value, str) else "wb"
        smart_write(o, value, mode=mode)


@cli.command("put")
def cli_put(
    key: str,
    value: Annotated[
        Optional[str],
        typer.Argument(..., help="Use this value instead of reading `-i` uri"),
    ] = None,
    i: Annotated[str, typer.Option("-i")] = "-",
):
    """
    Put content for a `key` to a store
    """
    with ErrorHandler():
        S = get_store(uri=state["uri"], use_pickle=state["pickle"])
        value = value or smart_read(i)
        S.put(key, value)


@cli.command("keys")
def cli_keys(
    prefix: Annotated[Optional[str], typer.Argument(..., help="Key prefix")] = None,
    o: Annotated[str, typer.Option("-o", help="Output uri")] = "-",
):
    """
    Iterate keys in given store
    """
    with ErrorHandler():
        S = get_store(uri=state["uri"], use_pickle=state["pickle"])
        keys = "\n".join(S.iterate_keys(prefix)) + "\n"
        smart_write(o, keys.encode())


@cli.command("io")
def cli_io(
    i: Annotated[str, typer.Option("-i", help="Input uri")] = "-",
    o: Annotated[str, typer.Option("-o", help="Output uri")] = "-",
):
    """
    Generic i/o wrapper using `anystore.io.smart_read` and
    `anystore.io.smart_write` which is wrapped around `fsspec`
    """
    with ErrorHandler():
        smart_write(o, smart_read(i))



from functools import cache
from typing import Generator, Optional, Union

from banal import ensure_dict
from sqlalchemy import (
    Column,
    DateTime,
    LargeBinary,
    MetaData,
    Table,
    Unicode,
    create_engine,
    func,
    insert,
    select,
    update,
)
from sqlalchemy.dialects.mysql import insert as mysql_insert
from sqlalchemy.dialects.postgresql import insert as psql_insert
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.engine import Connection, Engine

from anystore.exceptions import DoesNotExist
from anystore.settings import SqlSettings
from anystore.store.base import BaseStore
from anystore.types import Uri, Value


settings = SqlSettings()

Conn = Connection
Connish = Optional[Connection]


@cache
def get_engine(url: str, **kwargs) -> Engine:
    if "pool_size" not in kwargs:
        kwargs["pool_size"] = settings.pool_size
    return create_engine(url, **kwargs)


@cache
def get_metadata() -> MetaData:
    return MetaData()


Insert = Union[type(sqlite_insert), type(mysql_insert), type(psql_insert)]


def get_insert(engine: Engine) -> Insert:
    if engine.dialect.name == "sqlite":
        return sqlite_insert
    if engine.dialect.name == "mysql":
        return mysql_insert
    if engine.dialect.name in ("postgresql", "postgres"):
        return psql_insert
    raise RuntimeError("Unsupported database engine: %s" % engine.dialect.name)


def make_table(name: str, metadata: MetaData) -> Table:
    return Table(
        name,
        metadata,
        Column("key", Unicode(), primary_key=True, unique=True, index=True),
        Column("value", LargeBinary(), nullable=True),
        Column(
            "timestamp",
            DateTime(timezone=True),
            server_default=func.now(),
        ),
    )


class SqlStore(BaseStore):
    _engine: Engine | None = None
    _conn: Connection | None = None
    _insert: Insert | None = None
    _table: Table | None = None
    _sqlite: bool | None = True

    def __init__(self, **data):
        super().__init__(**data)
        backend_config = ensure_dict(self.backend_config)
        engine_kwargs = ensure_dict(backend_config.get("engine_kwargs"))
        metadata = get_metadata()
        engine = get_engine(self.uri, **engine_kwargs)
        table = backend_config.get("table") or settings.table
        table = make_table(table, metadata)
        metadata.create_all(engine, tables=[table], checkfirst=True)
        self._insert = get_insert(engine)
        self._table = table
        self._engine = engine
        self._conn = engine.connect()
        self._sqlite = "sqlite" in engine.name.lower()

    def _write(self, key: Uri, value: Value, **kwargs) -> None:
        key = str(key)
        # FIXME on conflict / on duplicate key
        exists = select(self._table).where(self._table.c.key == key)
        if self._conn.execute(exists).first():
            stmt = (
                update(self._table).where(self._table.c.key == key).values(value=value)
            )
        else:
            stmt = insert(self._table).values(key=key, value=value)
        self._conn.execute(stmt)
        self._conn.commit()

    def _read(self, key: Uri, raise_on_nonexist: bool | None = True, **kwargs) -> Value:
        key = str(key)
        stmt = select(self._table).where(self._table.c.key == key)
        res = self._conn.execute(stmt).first()
        if res:
            res = res[1]
            # mimic fs read mode:
            if kwargs.get("mode") == "r" and isinstance(res, bytes):
                res = res.decode()
            return res
        if raise_on_nonexist:
            raise DoesNotExist

    def _exists(self, key: Uri) -> bool:
        key = str(key)
        stmt = select(self._table).where(self._table.c.key == key)
        stmt = select(stmt.exists())
        for res in self._conn.execute(stmt).first():
            return res

    def _get_key_prefix(self) -> str:
        return ""

    def _iterate_keys(self, prefix: str | None = None) -> Generator[str, None, None]:
        table = self._table
        key = self.get_key(prefix or "")
        if not prefix:
            stmt = select(table.c.key)
        else:
            key = f"{key}%"
            stmt = select(table.c.key).where(table.c.key.like(key))
        with self._engine.connect() as conn:
            conn = conn.execution_options(stream_results=True)
            cursor = conn.execute(stmt)
            while rows := cursor.fetchmany(10_000):
                for row in rows:
                    yield row[0]



from anystore.store import get_store, Store
from anystore.decorators import anycache
from anystore.io import smart_read, smart_write, smart_stream

__all__ = [
    "get_store",
    "Store",
    "anycache",
    "smart_read",
    "smart_write",
    "smart_stream",
]


__version__ = "0.1.7"



"""
Store backend using any file-like location usable via `fsspec`
"""

from typing import Generator

import fsspec
from banal import ensure_dict

from anystore.io import smart_read, smart_stream, smart_write
from anystore.exceptions import DoesNotExist
from anystore.store.base import BaseStore
from anystore.types import Uri, Value


class Store(BaseStore):
    def _write(self, key: Uri, value: Value, **kwargs) -> None:
        return smart_write(key, value, **kwargs)

    def _read(
        self, key: Uri, raise_on_nonexist: bool | None = True, **kwargs
    ) -> Value | None:
        try:
            return smart_read(str(key), **kwargs)
        except FileNotFoundError:
            if raise_on_nonexist:
                raise DoesNotExist(f"Key does not exist: `{key}`")
            return None

    def _stream(
        self, key: Uri, raise_on_nonexist: bool | None = True, **kwargs
    ) -> Generator[Value, None, None]:
        try:
            yield from smart_stream(str(key), **kwargs)
        except FileNotFoundError:
            if raise_on_nonexist:
                raise DoesNotExist(f"Key does not exist: `{key}`")

    def _exists(self, key: Uri) -> bool:
        fs = fsspec.filesystem(self.scheme)
        return fs.exists(self.get_key(key))

    def _get_key_prefix(self) -> str:
        return str(self.uri).rstrip("/")

    def _iterate_keys(self, prefix: str | None = None) -> Generator[str, None, None]:
        path = self.get_key(prefix or "")
        mapper = fsspec.get_mapper(path, **ensure_dict(self.backend_config))
        for key in mapper.keys():
            if prefix:
                key = f"{prefix}/{key}"
            yield key



import orjson
from typing import Any, Literal, TypeAlias

import cloudpickle
from pydantic import BaseModel

from anystore.types import Model


Mode: TypeAlias = Literal["auto", "raw", "pickle", "json"]


def to_store(
    value: Any, serialization_mode: Mode | None = "auto", model: Model | None = None
) -> bytes:
    if model is not None:
        return value.model_dump_json().encode()
    mode = serialization_mode or "auto"
    if mode == "json":
        return orjson.dumps(value)
    if mode == "pickle":
        return cloudpickle.dumps(value)
    if mode == "raw":
        return value

    # auto
    if isinstance(value, BaseModel):
        return value.model_dump_json().encode()
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return value.encode()
    try:
        return orjson.dumps(value)
    except (orjson.JSONEncodeError, TypeError, ValueError):
        return cloudpickle.dumps(value)


def from_store(
    value: bytes,
    serialization_mode: Mode | None = "auto",
    model: Model | None = None,
) -> Any:
    if model is not None:
        data = orjson.loads(value)
        return model(**data)

    mode = serialization_mode or "auto"
    if mode == "raw":
        return value
    if mode == "pickle":
        return cloudpickle.loads(value)
    if mode == "json":
        return orjson.loads(value)

    # auto
    try:
        return orjson.loads(value)
    except (orjson.JSONDecodeError, TypeError, ValueError):
        try:
            return cloudpickle.loads(value)
        except Exception:
            if isinstance(value, bytes):
                try:
                    return value.decode()
                except UnicodeDecodeError:
                    pass
            return value
